# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import os
import sys
import shutil
from datetime import datetime
from pymic.util.parse_config import *
from pymic.net_run.agent_cls import ClassificationAgent
from pymic.net_run.agent_seg import SegmentationAgent
from pymic.net_run.semi_sup import SSLMethodDict
from pymic.net_run.weak_sup import WSLMethodDict
from pymic.net_run.noisy_label import NLLMethodDict
from pymic.net_run.self_sup import SelfSLSegAgent

def get_segmentation_agent(config, sup_type):
    assert(sup_type in ['fully_sup', 'semi_sup', 'self_sup', 'weak_sup', 'noisy_label'])
    if(sup_type == 'fully_sup'):
        logging.info("\n********** Fully Supervised Learning **********\n")
        agent = SegmentationAgent(config, 'train')
    elif(sup_type == 'semi_sup'):
        logging.info("\n********** Semi Supervised Learning **********\n")
        method = config['semi_supervised_learning']['method_name']
        agent = SSLMethodDict[method](config, 'train')
    elif(sup_type == 'weak_sup'):
        logging.info("\n********** Weakly Supervised Learning **********\n")
        method = config['weakly_supervised_learning']['method_name']
        agent = WSLMethodDict[method](config, 'train')
    elif(sup_type == 'noisy_label'):
        logging.info("\n********** Noisy Label Learning **********\n")
        method = config['noisy_label_learning']['method_name']
        agent = NLLMethodDict[method](config, 'train')
    elif(sup_type == 'self_sup'):
        logging.info("\n********** Self Supervised Learning **********\n")
        method = config['self_supervised_learning']['method_name']
        if(method == "custom"):
            pass
        elif(method == "model_genesis"):
            transforms = ['RandomFlip', 'LocalShuffling', 'NonLinearTransform', 'InOutPainting']
            genesis_cfg = {
                'randomflip_flip_depth': True,
                'randomflip_flip_height': True,
                'randomflip_flip_width': True,
                'localshuffling_probability': 0.5,
                'nonLineartransform_probability': 0.9,
                'inoutpainting_probability': 0.9,
                'inpainting_probability': 0.2
            }
            config['dataset']['train_transform'].extend(transforms)
            config['dataset']['valid_transform'].extend(transforms)
            config['dataset'].update(genesis_cfg)
            logging_config(config['dataset'])
        else:
            raise ValueError("The specified method {0:} is not implemented. ".format(method) + \
                         "Consider to set `self_sl_method = custom` and use customized" + \
                         " transforms for self-supervised learning.")
        agent = SelfSLSegAgent(config, 'train')
    else:
        raise ValueError("undefined supervision type: {0:}".format(sup_type))
    return agent

def main():
    """
    The main function for running a network for training.
    """
    if(len(sys.argv) < 2):
        print('Number of arguments should be 2. e.g.')
        print('   pymic_train config.cfg')
        exit()
    cfg_file = str(sys.argv[1])
    if(not os.path.isfile(cfg_file)):
        raise ValueError("The config file does not exist: " + cfg_file)
    config   = parse_config(cfg_file)
    config   = synchronize_config(config)
    log_dir  = config['training']['ckpt_save_dir']
    if(not os.path.exists(log_dir)):
        os.makedirs(log_dir, exist_ok=True)
    dst_cfg = cfg_file if "/" not in cfg_file else cfg_file.split("/")[-1]
    shutil.copy(cfg_file, log_dir + "/" + dst_cfg)
    if sys.version.startswith("3.9"):
        logging.basicConfig(filename=log_dir+"/log_train_{0:}.txt".format(str(datetime.now())[:-7]), 
                            level=logging.INFO, format='%(message)s', force=True) # for python 3.9
    else:
        logging.basicConfig(filename=log_dir+"/log_train_{0:}.txt".format(str(datetime.now())[:-7]), 
                            level=logging.INFO, format='%(message)s') # for python 3.6
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    logging_config(config)
    task     = config['dataset']['task_type']
    assert task in ['cls', 'cls_nexcl', 'seg']
    if(task == 'cls' or task == 'cls_nexcl'):
        agent = ClassificationAgent(config, 'train')
    else:
        sup_type = config['dataset'].get('supervise_type', 'fully_sup')
        agent = get_segmentation_agent(config, sup_type)
    agent.run()

if __name__ == "__main__":
    main()
    




# -*- coding: utf-8 -*-
from __future__ import print_function, division

import copy
# import logging
import math
import torch
import torch.nn as nn
from torch.nn import Dropout, Softmax, Linear, Conv2d, LayerNorm
from torch.nn.modules.utils import _pair

import numpy as np 
from scipy import ndimage
from os.path import join as pjoin
import pymic.net.net2d.trans2d.transunet_cfg as configs
from pymic.net.net2d.trans2d.transunet_resnet import ResNetV2


VIT_CONFIGS = {
    'ViT-B_16': configs.get_b16_config(),
    'ViT-B_32': configs.get_b32_config(),
    'ViT-L_16': configs.get_l16_config(),
    'ViT-L_32': configs.get_l32_config(),
    'ViT-H_14': configs.get_h14_config(),
    'R50-ViT-B_16': configs.get_r50_b16_config(),
    'R50-ViT-L_16': configs.get_r50_l16_config(),
    'testing': configs.get_testing(),
}

ATTENTION_Q = "MultiHeadDotProductAttention_1/query"
ATTENTION_K = "MultiHeadDotProductAttention_1/key"
ATTENTION_V = "MultiHeadDotProductAttention_1/value"
ATTENTION_OUT = "MultiHeadDotProductAttention_1/out"
FC_0 = "MlpBlock_3/Dense_0"
FC_1 = "MlpBlock_3/Dense_1"
ATTENTION_NORM = "LayerNorm_0"
MLP_NORM = "LayerNorm_2"


def np2th(weights, conv=False):
    """Possibly convert HWIO to OIHW."""
    if conv:
        weights = weights.transpose([3, 2, 0, 1])
    return torch.from_numpy(weights)


def swish(x):
    return x * torch.sigmoid(x)


ACT2FN = {"gelu": torch.nn.functional.gelu, "relu": torch.nn.functional.relu, "swish": swish}

class Attention(nn.Module):
    def __init__(self, config, vis):
        super(Attention, self).__init__()
        self.vis = vis
        self.num_attention_heads = config.transformer["num_heads"]
        self.attention_head_size = int(config.hidden_size / self.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = Linear(config.hidden_size, self.all_head_size)
        self.key = Linear(config.hidden_size, self.all_head_size)
        self.value = Linear(config.hidden_size, self.all_head_size)

        self.out = Linear(config.hidden_size, config.hidden_size)
        self.attn_dropout = Dropout(config.transformer["attention_dropout_rate"])
        self.proj_dropout = Dropout(config.transformer["attention_dropout_rate"])

        self.softmax = Softmax(dim=-1)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        attention_probs = self.softmax(attention_scores)
        weights = attention_probs if self.vis else None
        attention_probs = self.attn_dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        attention_output = self.out(context_layer)
        attention_output = self.proj_dropout(attention_output)
        return attention_output, weights

class Mlp(nn.Module):
    def __init__(self, config):
        super(Mlp, self).__init__()
        self.fc1 = Linear(config.hidden_size, config.transformer["mlp_dim"])
        self.fc2 = Linear(config.transformer["mlp_dim"], config.hidden_size)
        self.act_fn = ACT2FN["gelu"]
        self.dropout = Dropout(config.transformer["dropout_rate"])

        self._init_weights()

    def _init_weights(self):
        nn.init.xavier_uniform_(self.fc1.weight)
        nn.init.xavier_uniform_(self.fc2.weight)
        nn.init.normal_(self.fc1.bias, std=1e-6)
        nn.init.normal_(self.fc2.bias, std=1e-6)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act_fn(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.dropout(x)
        return x

class Embeddings(nn.Module):
    """Construct the embeddings from patch, position embeddings.
    """
    def __init__(self, config, img_size, in_channels=3):
        super(Embeddings, self).__init__()
        self.hybrid = None
        self.config = config
        img_size = _pair(img_size)

        if config.patches.get("grid") is not None:   # ResNet
            grid_size = config.patches["grid"]
            patch_size = (img_size[0] // 16 // grid_size[0], img_size[1] // 16 // grid_size[1])
            patch_size_real = (patch_size[0] * 16, patch_size[1] * 16)
            n_patches = (img_size[0] // patch_size_real[0]) * (img_size[1] // patch_size_real[1])  
            self.hybrid = True
        else:
            patch_size = _pair(config.patches["size"])
            n_patches = (img_size[0] // patch_size[0]) * (img_size[1] // patch_size[1])
            self.hybrid = False

        if self.hybrid:
            self.hybrid_model = ResNetV2(block_units=config.resnet.num_layers, width_factor=config.resnet.width_factor)
            in_channels = self.hybrid_model.width * 16
        self.patch_embeddings = Conv2d(in_channels=in_channels,
                                       out_channels=config.hidden_size,
                                       kernel_size=patch_size,
                                       stride=patch_size)
        self.position_embeddings = nn.Parameter(torch.zeros(1, n_patches, config.hidden_size))

        self.dropout = Dropout(config.transformer["dropout_rate"])


    def forward(self, x):
        if self.hybrid:
            x, features = self.hybrid_model(x)
        else:
            features = None
        x = self.patch_embeddings(x)  # (B, hidden. n_patches^(1/2), n_patches^(1/2))
        x = x.flatten(2)
        x = x.transpose(-1, -2)  # (B, n_patches, hidden)

        embeddings = x + self.position_embeddings
        embeddings = self.dropout(embeddings)
        return embeddings, features

class Block(nn.Module):
    def __init__(self, config, vis):
        super(Block, self).__init__()
        self.hidden_size = config.hidden_size
        self.attention_norm = LayerNorm(config.hidden_size, eps=1e-6)
        self.ffn_norm = LayerNorm(config.hidden_size, eps=1e-6)
        self.ffn = Mlp(config)
        self.attn = Attention(config, vis)

    def forward(self, x):
        h = x
        x = self.attention_norm(x)
        x, weights = self.attn(x)
        x = x + h

        h = x
        x = self.ffn_norm(x)
        x = self.ffn(x)
        x = x + h
        return x, weights

    def load_from(self, weights, n_block):
        ROOT = f"Transformer/encoderblock_{n_block}"
        with torch.no_grad():
            query_weight = np2th(weights[pjoin(ROOT, ATTENTION_Q, "kernel")]).view(self.hidden_size, self.hidden_size).t()
            key_weight = np2th(weights[pjoin(ROOT, ATTENTION_K, "kernel")]).view(self.hidden_size, self.hidden_size).t()
            value_weight = np2th(weights[pjoin(ROOT, ATTENTION_V, "kernel")]).view(self.hidden_size, self.hidden_size).t()
            out_weight = np2th(weights[pjoin(ROOT, ATTENTION_OUT, "kernel")]).view(self.hidden_size, self.hidden_size).t()

            query_bias = np2th(weights[pjoin(ROOT, ATTENTION_Q, "bias")]).view(-1)
            key_bias = np2th(weights[pjoin(ROOT, ATTENTION_K, "bias")]).view(-1)
            value_bias = np2th(weights[pjoin(ROOT, ATTENTION_V, "bias")]).view(-1)
            out_bias = np2th(weights[pjoin(ROOT, ATTENTION_OUT, "bias")]).view(-1)

            self.attn.query.weight.copy_(query_weight)
            self.attn.key.weight.copy_(key_weight)
            self.attn.value.weight.copy_(value_weight)
            self.attn.out.weight.copy_(out_weight)
            self.attn.query.bias.copy_(query_bias)
            self.attn.key.bias.copy_(key_bias)
            self.attn.value.bias.copy_(value_bias)
            self.attn.out.bias.copy_(out_bias)

            mlp_weight_0 = np2th(weights[pjoin(ROOT, FC_0, "kernel")]).t()
            mlp_weight_1 = np2th(weights[pjoin(ROOT, FC_1, "kernel")]).t()
            mlp_bias_0 = np2th(weights[pjoin(ROOT, FC_0, "bias")]).t()
            mlp_bias_1 = np2th(weights[pjoin(ROOT, FC_1, "bias")]).t()

            self.ffn.fc1.weight.copy_(mlp_weight_0)
            self.ffn.fc2.weight.copy_(mlp_weight_1)
            self.ffn.fc1.bias.copy_(mlp_bias_0)
            self.ffn.fc2.bias.copy_(mlp_bias_1)

            self.attention_norm.weight.copy_(np2th(weights[pjoin(ROOT, ATTENTION_NORM, "scale")]))
            self.attention_norm.bias.copy_(np2th(weights[pjoin(ROOT, ATTENTION_NORM, "bias")]))
            self.ffn_norm.weight.copy_(np2th(weights[pjoin(ROOT, MLP_NORM, "scale")]))
            self.ffn_norm.bias.copy_(np2th(weights[pjoin(ROOT, MLP_NORM, "bias")]))

class Encoder(nn.Module):
    def __init__(self, config, vis):
        super(Encoder, self).__init__()
        self.vis = vis
        self.layer = nn.ModuleList()
        self.encoder_norm = LayerNorm(config.hidden_size, eps=1e-6)
        for _ in range(config.transformer["num_layers"]):
            layer = Block(config, vis)
            self.layer.append(copy.deepcopy(layer))

    def forward(self, hidden_states):
        attn_weights = []
        for layer_block in self.layer:
            hidden_states, weights = layer_block(hidden_states)
            if self.vis:
                attn_weights.append(weights)
        encoded = self.encoder_norm(hidden_states)
        return encoded, attn_weights

class Transformer(nn.Module):
    def __init__(self, config, img_size, vis):
        super(Transformer, self).__init__()
        self.embeddings = Embeddings(config, img_size=img_size)
        self.encoder = Encoder(config, vis)

    def forward(self, input_ids):
        embedding_output, features = self.embeddings(input_ids)
        encoded, attn_weights = self.encoder(embedding_output)  # (B, n_patch, hidden)
        return encoded, attn_weights, features

class Conv2dReLU(nn.Sequential):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size,
            padding=0,
            stride=1,
            use_batchnorm=True,
    ):
        conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=not (use_batchnorm),
        )
        relu = nn.ReLU(inplace=True)

        bn = nn.BatchNorm2d(out_channels)

        super(Conv2dReLU, self).__init__(conv, bn, relu)


class DecoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            skip_channels=0,
            use_batchnorm=True,
    ):
        super().__init__()
        self.conv1 = Conv2dReLU(
            in_channels + skip_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.conv2 = Conv2dReLU(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.up = nn.UpsamplingBilinear2d(scale_factor=2)

    def forward(self, x, skip=None):
        x = self.up(x)
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class SegmentationHead(nn.Sequential):

    def __init__(self, in_channels, out_channels, kernel_size=3, upsampling=1):
        conv2d = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2)
        upsampling = nn.UpsamplingBilinear2d(scale_factor=upsampling) if upsampling > 1 else nn.Identity()
        super().__init__(conv2d, upsampling)


class DecoderCup(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        head_channels = 512
        self.conv_more = Conv2dReLU(
            config.hidden_size,
            head_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=True,
        )
        decoder_channels = config.decoder_channels
        in_channels = [head_channels] + list(decoder_channels[:-1])
        out_channels = decoder_channels

        if self.config.n_skip != 0:
            skip_channels = self.config.skip_channels
            for i in range(4-self.config.n_skip):  # re-select the skip channels according to n_skip
                skip_channels[3-i]=0

        else:
            skip_channels=[0,0,0,0]

        blocks = [
            DecoderBlock(in_ch, out_ch, sk_ch) for in_ch, out_ch, sk_ch in zip(in_channels, out_channels, skip_channels)
        ]
        self.blocks = nn.ModuleList(blocks)

    def forward(self, hidden_states, features=None):
        B, n_patch, hidden = hidden_states.size()  # reshape from (B, n_patch, hidden) to (B, h, w, hidden)
        h, w = int(np.sqrt(n_patch)), int(np.sqrt(n_patch))
        x = hidden_states.permute(0, 2, 1)
        x = x.contiguous().view(B, hidden, h, w)
        x = self.conv_more(x)
        for i, decoder_block in enumerate(self.blocks):
            if features is not None:
                skip = features[i] if (i < self.config.n_skip) else None
            else:
                skip = None
            x = decoder_block(x, skip=skip)
        return x

class TransUNet(nn.Module):
    def __init__(self, params):
        super(TransUNet, self).__init__()
        vit_name    = params.get("vit_name", 'R50-ViT-B_16')
        img_size    = params['img_size']
        vis         = params.get("vis", False)
        self.config = VIT_CONFIGS[vit_name]
        self.num_classes = params['class_num']
        self.zero_head   = params.get("zero_head",  False)

        self.classifier = self.config.classifier
        self.transformer = Transformer(self.config, img_size, vis)
        self.decoder = DecoderCup(self.config)
        self.segmentation_head = SegmentationHead(
            in_channels=self.config['decoder_channels'][-1],
            out_channels=self.num_classes,
            kernel_size=3,
        )

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)

        if x.size()[1] == 1:
            x = x.repeat(1,3,1,1)
        x, attn_weights, features = self.transformer(x)  # (B, n_patch, hidden)
        x = self.decoder(x, features)
        logits = self.segmentation_head(x)

        if(len(x_shape) == 5):
            new_shape = [N, D] + list(logits.shape)[1:]
            logits = torch.reshape(logits, new_shape)
            logits = torch.transpose(logits, 1, 2)

        return logits

    def load_from(self, weights):
        with torch.no_grad():

            res_weight = weights
            self.transformer.embeddings.patch_embeddings.weight.copy_(np2th(weights["embedding/kernel"], conv=True))
            self.transformer.embeddings.patch_embeddings.bias.copy_(np2th(weights["embedding/bias"]))

            self.transformer.encoder.encoder_norm.weight.copy_(np2th(weights["Transformer/encoder_norm/scale"]))
            self.transformer.encoder.encoder_norm.bias.copy_(np2th(weights["Transformer/encoder_norm/bias"]))

            posemb = np2th(weights["Transformer/posembed_input/pos_embedding"])

            posemb_new = self.transformer.embeddings.position_embeddings
            if posemb.size() == posemb_new.size():
                self.transformer.embeddings.position_embeddings.copy_(posemb)
            elif posemb.size()[1]-1 == posemb_new.size()[1]:
                posemb = posemb[:, 1:]
                self.transformer.embeddings.position_embeddings.copy_(posemb)
            else:
                # logger.info("load_pretrained: resized variant: %s to %s" % (posemb.size(), posemb_new.size()))
                ntok_new = posemb_new.size(1)
                if self.classifier == "seg":
                    _, posemb_grid = posemb[:, :1], posemb[0, 1:]
                gs_old = int(np.sqrt(len(posemb_grid)))
                gs_new = int(np.sqrt(ntok_new))
                print('load_pretrained: grid-size from %s to %s' % (gs_old, gs_new))
                posemb_grid = posemb_grid.reshape(gs_old, gs_old, -1)
                zoom = (gs_new / gs_old, gs_new / gs_old, 1)
                posemb_grid = ndimage.zoom(posemb_grid, zoom, order=1)  # th2np
                posemb_grid = posemb_grid.reshape(1, gs_new * gs_new, -1)
                posemb = posemb_grid
                self.transformer.embeddings.position_embeddings.copy_(np2th(posemb))

            # Encoder whole
            for bname, block in self.transformer.encoder.named_children():
                for uname, unit in block.named_children():
                    unit.load_from(weights, n_block=uname)

            if self.transformer.embeddings.hybrid:
                self.transformer.embeddings.hybrid_model.root.conv.weight.copy_(np2th(res_weight["conv_root/kernel"], conv=True))
                gn_weight = np2th(res_weight["gn_root/scale"]).view(-1)
                gn_bias = np2th(res_weight["gn_root/bias"]).view(-1)
                self.transformer.embeddings.hybrid_model.root.gn.weight.copy_(gn_weight)
                self.transformer.embeddings.hybrid_model.root.gn.bias.copy_(gn_bias)

                for bname, block in self.transformer.embeddings.hybrid_model.body.named_children():
                    for uname, unit in block.named_children():
                        unit.load_from(res_weight, n_block=bname, n_unit=uname)

if __name__ == "__main__":
    params = {'img_size': [256, 256],
              'class_num': 2}
    net = TransUNet(params)
    net.double()

    x  = np.random.rand(4, 1, 256, 256)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = net(xt)
    print(len(y.size()))
    y = y.detach().numpy()
    print(y.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np 
from pymic.loss.seg.util import reshape_tensor_to_2D
from pymic.loss.seg.abstract import AbstractSegLoss

class EntropyLoss(AbstractSegLoss):
    """
    Entropy Minimization for segmentation tasks.
    The parameters should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (bool) Apply softmax to the prediction of network or not. 
    """
    def __init__(self, params = None):
        super(EntropyLoss, self).__init__(params)
        
    def forward(self, loss_input_dict):
        """
        Forward pass for calculating the loss.
        The arguments should be written in the `loss_input_dict` dictionary, 
        and it has the following fields:

        :param `prediction`: (tensor) Prediction of a network, with the 
            shape of [N, C, D, H, W] or [N, C, H, W].

        :return: Loss function value.
        """
        predict = loss_input_dict['prediction']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        C = list(predict.shape)[1]
        entropy = torch.sum(-predict*torch.log(predict), dim=1) / np.log(C)
        avg_ent = torch.mean(entropy)
        return avg_ent
    
class TotalVariationLoss(AbstractSegLoss):
    """
    Total Variation Loss for segmentation tasks.
    The parameters should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (bool) Apply softmax to the prediction of network or not. 
    """
    def __init__(self, params = None):
        super(TotalVariationLoss, self).__init__(params)
        
    def forward(self, loss_input_dict):
        """
        Forward pass for calculating the loss.
        The arguments should be written in the `loss_input_dict` dictionary, 
        and it has the following fields:

        :param `prediction`: (tensor) Prediction of a network, with the 
            shape of [N, C, D, H, W] or [N, C, H, W].

        :return: Loss function value.
        """
        predict = loss_input_dict['prediction']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        dim = list(predict.shape)[2:]
        if(dim == 2):
            pred_min = -1 * nn.functional.max_pool2d(-1*predict, (3, 3), 1, 1)
            pred_max = nn.functional.max_pool2d(pred_min, (3, 3), 1, 1)
        else:
            pred_min = -1 * nn.functional.max_pool3d(-1*predict, (3, 3, 3), 1, 1)
            pred_max = nn.functional.max_pool3d(pred_min, (3, 3, 3), 1, 1)            
        contour = torch.relu(pred_max - pred_min)
        length  = torch.mean(contour)
        return length



# -*- coding: utf-8 -*-
"""
Functions for ramping hyperparameters up or down.

Each function takes the current training step or epoch, and the
ramp length (start and end step or epoch), and returns a multiplier between
0 and 1.
"""
from __future__ import print_function, division
import numpy as np

def get_rampup_ratio(i, start, end, mode = "linear"):
    """
    Obtain the rampup ratio.

    :param i: (int) The current iteration.
    :param start: (int) The start iteration.
    :param end: (int) The end itertation.
    :param mode: (str) Valid values are {`linear`, `sigmoid`, `cosine`}.
    """
    i = np.clip(i, start, end)
    if(mode == "linear"):
        rampup = (i - start) / (end - start)
    elif(mode == "sigmoid"):
        phase  = 1.0 - (i - start) / (end - start)
        rampup = float(np.exp(-5.0 * phase * phase))
    elif(mode == "cosine"):
        phase  = 1.0 - (i - start) / (end - start)
        rampup = float(.5 * (np.cos(np.pi * phase) + 1))
    else:
        raise ValueError("Undefined rampup mode {0:}".format(mode))
    return rampup


def get_rampdown_ratio(i, start, end, mode = "linear"):
    """
    Obtain the rampdown ratio.

    :param i: (int) The current iteration.
    :param start: (int) The start iteration.
    :param end: (int) The end itertation.
    :param mode: (str) Valid values are {`linear`, `sigmoid`, `cosine`}.
    """
    i = np.clip(i, start, end)
    if(mode == "linear"):
        rampdown = 1.0 - (i - start) / (end - start)
    elif(mode == "sigmoid"):
        phase = (i - start) / (end - start)
        rampdown = float(np.exp(-5.0 * phase * phase))
    elif(mode == "cosine"):
        phase = (i - start) / (end - start)
        rampdown = float(.5 * (np.cos(np.pi * phase) + 1))
    else:
        raise ValueError("Undefined rampup mode {0:}".format(mode))
    return rampdown

    


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from pymic.transform.intensity import  *
from pymic.transform.flip import *
from pymic.transform.pad import *
from pymic.transform.rotate import *
from pymic.transform.rescale import *
from pymic.transform.transpose import *
from pymic.transform.threshold import * 
from pymic.transform.normalize import *
from pymic.transform.crop import *
from pymic.transform.label_convert import * 
from pymic.transform.trans_dict import TransformDict


import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class NormalizeWithMeanStd(AbstractTransform):
    """
    Normalize the image based on mean and std. The image should have a shape
    of [C, D, H, W] or [C, H, W].  

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `NormalizeWithMeanStd_channels`: (list/tuple or None) 
        A list or tuple of int for specifying the channels. 
        If None, the transform operates on all the channels.
    :param `NormalizeWithMeanStd_mean`: (list/tuple or None) 
        The mean values along each specified channel.
        If None, the mean values are calculated automatically.
    :param `NormalizeWithMeanStd_std`: (list/tuple or None) 
        The std values along each specified channel.
        If None, the std values are calculated automatically.
    :param `NormalizeWithMeanStd_ignore_non_positive`: (optional, bool) 
        Only used when mean and std are not given. Default is False.
        If True, calculate mean and std in the positive region for normalization,
        and set non-positive region to random. If False, calculate
        the mean and std values in the entire image region. 
    :param `NormalizeWithMeanStd_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(NormalizeWithMeanStd, self).__init__(params)
        self.chns = params['NormalizeWithMeanStd_channels'.lower()]
        self.mean = params.get('NormalizeWithMeanStd_mean'.lower(), None)
        self.std  = params.get('NormalizeWithMeanStd_std'.lower(), None)
        self.ingore_np = params.get('NormalizeWithMeanStd_ignore_non_positive'.lower(), False)
        self.inverse = params.get('NormalizeWithMeanStd_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        if(self.mean is None):
            self.mean = [None] * len(chns)
            self.std  = [None] * len(chns)

        for i in range(len(chns)):
            chn = chns[i]
            chn_mean, chn_std = self.mean[i], self.std[i]
            if(chn_mean is None):
                if(self.ingore_np):
                    pixels = image[chn][image[chn] > 0]
                    chn_mean, chn_std = pixels.mean(), pixels.std() 
                else:
                    chn_mean, chn_std = image[chn].mean(), image[chn].std()
    
            chn_norm = (image[chn] - chn_mean)/chn_std

            if(self.ingore_np):
                chn_random = np.random.normal(0, 1, size = chn_norm.shape)
                chn_norm[image[chn] <= 0] = chn_random[image[chn] <= 0]
            image[chn] = chn_norm
        sample['image'] = image
        return sample


class NormalizeWithMinMax(AbstractTransform):
    """Nomralize the image to [0, 1]. The shape should be [C, D, H, W] or [C, H, W].

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `NormalizeWithMinMax_channels`: (list/tuple or None) 
        A list or tuple of int for specifying the channels. 
        If None, the transform operates on all the channels.
    :param `NormalizeWithMinMax_threshold_lower`: (list/tuple or None) 
        The min values along each specified channel.
        If None, the min values are calculated automatically.
    :param `NormalizeWithMinMax_threshold_upper`: (list/tuple or None) 
        The max values along each specified channel.
        If None, the max values are calculated automatically.
    :param `NormalizeWithMinMax_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(NormalizeWithMinMax, self).__init__(params)
        self.chns = params['NormalizeWithMinMax_channels'.lower()]
        self.thred_lower = params['NormalizeWithMinMax_threshold_lower'.lower()]
        self.thred_upper = params['NormalizeWithMinMax_threshold_upper'.lower()]
        self.inverse = params.get('NormalizeWithMinMax_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        for i in range(len(chns)):
            chn = chns[i]
            img_chn = image[chn]
            v0, v1 = img_chn.min(), img_chn.max()
            if(self.thred_lower is not None) and  (self.thred_lower[i] is not None):
                v0 = self.thred_lower[i]
            if(self.thred_upper is not None) and  (self.thred_upper[i] is not None):
                v1 = self.thred_upper[i]

            img_chn[img_chn < v0] = v0
            img_chn[img_chn > v1] = v1
            img_chn = (img_chn - v0) / (v1 - v0)
            image[chn] = img_chn
        sample['image'] = image
        return sample

class NormalizeWithPercentiles(AbstractTransform):
    """Nomralize the image to [0, 1] with percentiles for given channels.
    The shape should be [C, D, H, W] or [C, H, W].
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `NormalizeWithPercentiles_channels`: (list/tuple or None) 
        A list or tuple of int for specifying the channels. 
        If None, the transform operates on all the channels.
    :param `NormalizeWithPercentiles_percentile_lower`: (float) 
        The min percentile, which must be between 0 and 100 inclusive.
    :param `NormalizeWithPercentiles_percentile_upper`: (float) 
        The max percentile, which must be between 0 and 100 inclusive.
    :param `NormalizeWithMinMax_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(NormalizeWithPercentiles, self).__init__(params)
        self.chns = params['NormalizeWithPercentiles_channels'.lower()]
        self.percent_lower = params['NormalizeWithPercentiles_percentile_lower'.lower()]
        self.percent_upper = params['NormalizeWithPercentiles_percentile_upper'.lower()]
        self.inverse = params.get('NormalizeWithPercentiles_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        for i in range(len(chns)):
            chn = chns[i]
            img_chn = image[chn]
            v0 = np.percentile(img_chn, self.percent_lower)
            v1 = np.percentile(img_chn, self.percent_upper)

            img_chn[img_chn < v0] = v0
            img_chn[img_chn > v1] = v1
            img_chn = (img_chn - v0) / (v1 - v0)
            image[chn] = img_chn
        sample['image'] = image
        return sample


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np
from torch.nn.functional import interpolate

class ConvBlock(nn.Module):
    """
    Two 3D convolution layers with batch norm and leaky relu.
    Droput is used between the two convolution layers.
    
    :param in_channels: (int) Input channel number.
    :param out_channels: (int) Output channel number.
    :param dropout_p: (int) Dropout probability.
    """
    def __init__(self, in_channels, out_channels, dropout_p):
        super(ConvBlock, self).__init__()
        self.conv_conv = nn.Sequential(
            nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(),
            nn.Dropout(dropout_p),
            nn.Conv3d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU()
        )
       
    def forward(self, x):
        return self.conv_conv(x)

class DownBlock(nn.Module):
    """
    3D downsampling followed by ConvBlock

    :param in_channels: (int) Input channel number.
    :param out_channels: (int) Output channel number.
    :param dropout_p: (int) Dropout probability.
    """
    def __init__(self, in_channels, out_channels, dropout_p):
        super(DownBlock, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool3d(2),
            ConvBlock(in_channels, out_channels, dropout_p)
        )

    def forward(self, x):
        return self.maxpool_conv(x)

class UpBlock(nn.Module):
    """
    3D upsampling followed by ConvBlock
    
    :param in_channels1: (int) Channel number of high-level features.
    :param in_channels2: (int) Channel number of low-level features.
    :param out_channels: (int) Output channel number.
    :param dropout_p: (int) Dropout probability.
    :param trilinear: (bool) Use trilinear for up-sampling (by default).
        If False, deconvolution is used for up-sampling. 
    """
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p,
                 trilinear=True):
        super(UpBlock, self).__init__()
        self.trilinear = trilinear
        if trilinear:
            self.conv1x1 = nn.Conv3d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBlock(in_channels2 * 2, out_channels, dropout_p)

    def forward(self, x1, x2):
        if self.trilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)

class Encoder(nn.Module):
    """
    Encoder of 3D UNet.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    """
    def __init__(self, params):
        super(Encoder, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        self.in_conv= ConvBlock(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        if(len(self.ft_chns) == 5):
            self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])

    def forward(self, x):
        x0 = self.in_conv(x)
        x1 = self.down1(x0)
        x2 = self.down2(x1)
        x3 = self.down3(x2)
        output = [x0, x1, x2, x3]
        if(len(self.ft_chns) == 5):
          x4 = self.down4(x3)
          output.append(x4)
        return output

class Decoder(nn.Module):
    """
    Decoder of 3D UNet.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using bilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    :param multiscale_pred: (bool) Get multi-scale prediction.
    """
    def __init__(self, params):
        super(Decoder, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        self.n_class   = self.params['class_num']
        self.trilinear = self.params.get('trilinear', True)
        self.mul_pred  = self.params.get('multiscale_pred', False)

        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        if(len(self.ft_chns) == 5):
            self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], self.dropout[3], self.trilinear) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], self.dropout[2], self.trilinear) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], self.dropout[1], self.trilinear) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], self.dropout[0], self.trilinear) 
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, kernel_size = 1)
        if(self.mul_pred):
            self.out_conv1 = nn.Conv3d(self.ft_chns[1], self.n_class, kernel_size = 1)
            self.out_conv2 = nn.Conv3d(self.ft_chns[2], self.n_class, kernel_size = 1)
            self.out_conv3 = nn.Conv3d(self.ft_chns[3], self.n_class, kernel_size = 1)

    def forward(self, x):
        if(len(self.ft_chns) == 5):
            assert(len(x) == 5)
            x0, x1, x2, x3, x4 = x 
            x_d3 = self.up1(x4, x3)
        else:
            assert(len(x) == 4)
            x0, x1, x2, x3 = x 
            x_d3 = x3
        x_d2 = self.up2(x_d3, x2)
        x_d1 = self.up3(x_d2, x1)
        x_d0 = self.up4(x_d1, x0)
        output = self.out_conv(x_d0)
        if(self.mul_pred):
            output1 = self.out_conv1(x_d1)
            output2 = self.out_conv2(x_d2)
            output3 = self.out_conv3(x_d3)
            output = [output, output1, output2, output3]
        return output

class UNet3D(nn.Module):
    """
    An implementation of the U-Net.
        
    * Reference: Özgün Çiçek, Ahmed Abdulkadir, Soeren S. Lienkamp, Thomas Brox, Olaf Ronneberger:
      3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation. 
      `MICCAI (2) 2016: 424-432. <https://link.springer.com/chapter/10.1007/978-3-319-46723-8_49>`_
    
    Note that there are some modifications from the original paper, such as
    the use of batch normalization, dropout, leaky relu and deep supervision.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using trilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    :param multiscale_pred: (bool) Get multi-scale prediction.
    """
    def __init__(self, params):
        super(UNet3D, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        self.n_class   = self.params['class_num']
        self.trilinear = self.params['trilinear']
        self.mul_pred  = self.params['multiscale_pred']
        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        self.in_conv= ConvBlock(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        if(len(self.ft_chns) == 5):
          self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])
          self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], 
               dropout_p = self.dropout[3], trilinear=self.trilinear) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], 
               dropout_p = self.dropout[2], trilinear=self.trilinear) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], 
               dropout_p = self.dropout[1], trilinear=self.trilinear) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], 
               dropout_p = self.dropout[0], trilinear=self.trilinear) 
    
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, kernel_size = 1)
        if(self.mul_pred):
            self.out_conv1 = nn.Conv3d(self.ft_chns[1], self.n_class, kernel_size = 1)
            self.out_conv2 = nn.Conv3d(self.ft_chns[2], self.n_class, kernel_size = 1)
            self.out_conv3 = nn.Conv3d(self.ft_chns[3], self.n_class, kernel_size = 1)

    def forward(self, x):
        x0 = self.in_conv(x)
        x1 = self.down1(x0)
        x2 = self.down2(x1)
        x3 = self.down3(x2)
        if(len(self.ft_chns) == 5):
          x4 = self.down4(x3)
          x_d3 = self.up1(x4, x3)
        else:
          x_d3 = x3
        x_d2 = self.up2(x_d3, x2)
        x_d1 = self.up3(x_d2, x1)
        x_d0 = self.up4(x_d1, x0)
        output = self.out_conv(x_d0)
        if(self.mul_pred):
            output1 = self.out_conv1(x_d1)
            output2 = self.out_conv2(x_d2)
            output3 = self.out_conv3(x_d3)
            output = [output, output1, output2, output3]
        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'class_num': 2,
              'feature_chns':[2, 8, 32, 64],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True,
              'multiscale_pred': False}
    Net = UNet3D(params)
    Net = Net.double()

    x  = np.random.rand(4, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    y = y.detach().numpy()
    print(y.shape)



from __future__ import absolute_import
from pymic.net_run.semi_sup.ssl_abstract import SSLSegAgent
from pymic.net_run.semi_sup.ssl_em import SSLEntropyMinimization
from pymic.net_run.semi_sup.ssl_mt import SSLMeanTeacher
from pymic.net_run.semi_sup.ssl_uamt import SSLUncertaintyAwareMeanTeacher
from pymic.net_run.semi_sup.ssl_cct import SSLCCT
from pymic.net_run.semi_sup.ssl_cps import SSLCPS
from pymic.net_run.semi_sup.ssl_urpc import SSLURPC


SSLMethodDict = {'EntropyMinimization': SSLEntropyMinimization,
    'MeanTeacher': SSLMeanTeacher,
    'UAMT': SSLUncertaintyAwareMeanTeacher,
    'CCT': SSLCCT,
    'CPS': SSLCPS,
    'URPC': SSLURPC}


# -*- coding: utf-8 -*-
"""
Built-in loss functions for segmentation. 
The following are for fully supervised learning, or learnig from noisy labels:

* CrossEntropyLoss :mod:`pymic.loss.seg.ce.CrossEntropyLoss`
* GeneralizedCELoss :mod:`pymic.loss.seg.ce.GeneralizedCELoss`
* DiceLoss :mod:`pymic.loss.seg.dice.DiceLoss`
* FocalDiceLoss :mod:`pymic.loss.seg.dice.FocalDiceLoss`
* NoiseRobustDiceLoss :mod:`pymic.loss.seg.dice.NoiseRobustDiceLoss`
* ExpLogLoss :mod:`pymic.loss.seg.exp_log.ExpLogLoss`
* MAELoss  :mod:`pymic.loss.seg.mse.MAELoss`
* MSELoss  :mod:`pymic.loss.seg.mse.MSELoss`
* SLSRLoss :mod:`pymic.loss.seg.slsr.SLSRLoss`

The following are for semi-supervised or weakly supervised learning:

* EntropyLoss :mod:`pymic.loss.seg.ssl.EntropyLoss`
* GatedCRFLoss: :mod:`pymic.loss.seg.gatedcrf.GatedCRFLoss`
* MumfordShahLoss  :mod:`pymic.loss.seg.mumford_shah.MumfordShahLoss`
* TotalVariationLoss :mod:`pymic.loss.seg.ssl.TotalVariationLoss`
"""
from __future__ import print_function, division
import torch.nn as nn 
from pymic.loss.seg.ce import CrossEntropyLoss, GeneralizedCELoss
from pymic.loss.seg.dice import DiceLoss, FocalDiceLoss, NoiseRobustDiceLoss
from pymic.loss.seg.exp_log import ExpLogLoss
from pymic.loss.seg.mse import MSELoss, MAELoss
from pymic.loss.seg.slsr import SLSRLoss

SegLossDict = {
    'CrossEntropyLoss': CrossEntropyLoss,
    'GeneralizedCELoss': GeneralizedCELoss,
    'DiceLoss': DiceLoss,
    'FocalDiceLoss': FocalDiceLoss,
    'NoiseRobustDiceLoss': NoiseRobustDiceLoss,
    'ExpLogLoss': ExpLogLoss,
    'MAELoss': MAELoss,
    'MSELoss': MSELoss,
    'SLSRLoss': SLSRLoss
    }




from __future__ import absolute_import
from pymic.net_run.noisy_label.nll_co_teaching import NLLCoTeaching
from pymic.net_run.noisy_label.nll_trinet import NLLTriNet
from pymic.net_run.noisy_label.nll_dast import NLLDAST

NLLMethodDict = {'CoTeaching': NLLCoTeaching,
    "TriNet": NLLTriNet,
    "DAST": NLLDAST}


# -*- coding: utf-8 -*-
"""
Built-in loss functions for classification. 

* CrossEntropyLoss :mod:`pymic.loss.cls.basic.CrossEntropyLoss`
* SigmoidCELoss :mod:`pymic.loss.cls.basic.SigmoidCELoss`
* L1Loss :mod:`pymic.loss.cls.basic.L1Loss`
* MSELoss :mod:`pymic.loss.cls.basic.MSELoss`
* NLLLoss :mod:`pymic.loss.cls.basic.NLLLoss`

"""
from __future__ import print_function, division
from pymic.loss.cls.basic import *

PyMICClsLossDict = {"CrossEntropyLoss": CrossEntropyLoss,
    "SigmoidCELoss": SigmoidCELoss,
    "L1Loss":  L1Loss,
    "MSELoss": MSELoss,
    "NLLLoss": NLLLoss}



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch.nn as nn
from torch.nn.functional import interpolate
from pymic.loss.seg.abstract import AbstractSegLoss

def match_prediction_and_gt_shape(pred, gt, mode = 0):
    pred_shape = list(pred.shape)
    gt_shape   = list(gt.shape)
    dim = len(pred_shape) - 2
    shape_match = False 
    if(dim == 2):
        if(pred_shape[-1] == gt_shape[-1] and pred_shape[-2] == gt_shape[-2]):
            shape_match = True
    else:
        if(pred_shape[-1] == gt_shape[-1] and pred_shape[-2] == gt_shape[-2]
           and pred_shape[-3] == gt_shape[-3]):
            shape_match = True
    if(shape_match):
        return pred, gt 
    
    interp_mode = 'bilinear' if dim == 2 else 'trilinear'
    if(mode == 0):
        pred_new = interpolate(pred, gt_shape[2:], mode = interp_mode)
        gt_new   = gt  
    elif(mode == 1):
        pred_new = pred
        gt_new   = interpolate(gt, pred_shape[2:], mode = interp_mode)
    elif(mode == 2):
        pred_new = pred
        if(dim == 2):
            avg_pool = nn.AdaptiveAvgPool2d(pred_shape[-2:])
        else:
            avg_pool = nn.AdaptiveAvgPool3d(pred_shape[-3:])
        gt_new = avg_pool(gt)
    else:
        raise ValueError("mode shoud be 0, 1 or 2, but {0:} was given".format(mode))
    return pred_new, gt_new
            

class DeepSuperviseLoss(AbstractSegLoss):
    '''
    Combine deep supervision with a basic loss function.  
    Arguments should be provided in the `params` dictionary, and it has the 
    following fields:

    :param `loss_softmax`: (optional, bool) 
        Apply softmax to the prediction of network or not. Default is True.
    :param `base_loss`: (nn.Module) The basic function used for each scale.
    :param `deep_supervise_weight`: (list) A list of weight for each deep supervision scale. 
    :param `deep_supervise_model`: (int) Mode for deep supervision when the prediction
        has a smaller shape than the ground truth. 0: upsample the prediction to the size 
        of the ground truth. 1: downsample the ground truth to the size of the prediction
        via interpolation. 2: downsample the ground truth via adaptive average pooling.

    '''
    def __init__(self, params):
        super(DeepSuperviseLoss, self).__init__(params)
        self.base_loss       = params['base_loss']
        self.deep_sup_weight = params.get('deep_supervise_weight', None)
        self.deep_sup_mode   = params.get('deep_supervise_mode', 0)

    def forward(self, loss_input_dict):
        pred = loss_input_dict['prediction']
        gt   = loss_input_dict['ground_truth']
        if(not isinstance(pred, (list,tuple))):
            raise ValueError("""For deep supervision, the prediction should
                be a list or a tuple""")
        pred_num = len(pred)
        if(self.deep_sup_weight is None):
            self.deep_sup_weight = [1.0] * pred_num
        else:
            assert(pred_num == len(self.deep_sup_weight))
        loss_sum, weight_sum  = 0.0, 0.0
        for i in range(pred_num):
            pred_i, gt_i = match_prediction_and_gt_shape(pred[i], gt, self.deep_sup_mode)
            loss_input_dict['prediction']   = pred_i
            loss_input_dict['ground_truth'] = gt_i
            temp_loss   = self.base_loss(loss_input_dict)
            loss_sum   += temp_loss * self.deep_sup_weight[i]
            weight_sum += self.deep_sup_weight[i]
        loss = loss_sum/weight_sum
        return loss


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import os
import numpy as np
import SimpleITK as sitk
from PIL import Image

def load_nifty_volume_as_4d_array(filename):
    """
    Read a nifty image and return a dictionay storing data array, origin, 
    spacing and direction.\n
    output['data_array'] 4D array with shape [C, D, H, W];\n
    output['spacing']    A list of spacing in z, y, x axis;\n
    output['direction']  A 3x3 matrix for direction.

    :param filename: (str) The input file name
    :return: A dictionay storing data array, origin, spacing and direction.
    """
    img_obj    = sitk.ReadImage(filename)
    data_array = sitk.GetArrayFromImage(img_obj)
    origin     = img_obj.GetOrigin()
    spacing    = img_obj.GetSpacing()
    direction  = img_obj.GetDirection()
    shape = data_array.shape
    if(len(shape) == 4):
        assert(shape[3] == 1) 
    elif(len(shape) == 3):
        data_array = np.expand_dims(data_array, axis = 0)
    else:
        raise ValueError("unsupported image dim: {0:}".format(len(shape)))
    output = {}
    output['data_array'] = data_array
    output['origin']     = origin
    output['spacing']    = (spacing[2], spacing[1], spacing[0])
    output['direction']  = direction
    return output

def load_rgb_image_as_3d_array(filename):
    """
    Read an RGB image and return a dictionay storing data array, origin, 
    spacing and direction. \n
    output['data_array'] 3D array with shape [D, H, W]; \n
    output['spacing']    a list of spacing in z, y, x axis;  \n
    output['direction']  a 3x3 matrix for direction.

    :param filename: (str) The input file name
    :return: A dictionay storing data array, origin, spacing and direction.
    """
    image = np.asarray(Image.open(filename))
    image_shape = image.shape
    image_dim   = len(image_shape)
    assert(image_dim == 2 or image_dim == 3)
    if(image_dim == 2):
        image = np.expand_dims(image, axis = 0)
    else:
        # transpose rgb image from [H, W, C] to [C, H, W]
        assert(image_shape[2] == 3 or image_shape[2] == 4)
        if(image_shape[2] == 4):
            image = image[:, :, range(3)]
        image = np.transpose(image, axes = [2, 0, 1])
    output = {}
    output['data_array'] = image
    output['origin']     = (0, 0)
    output['spacing']    = (1.0, 1.0)
    output['direction']  = 0
    return output

def load_image_as_nd_array(image_name):
    """
    Load an image and return a 4D array with shape [C, D, H, W], 
    or 3D array with shape [C, H, W].

    :param filename: (str) The input file name
    :return: A dictionay storing data array, origin, spacing and direction.
    """
    if (image_name.endswith(".nii.gz") or image_name.endswith(".nii") or
        image_name.endswith(".mha")):
        image_dict = load_nifty_volume_as_4d_array(image_name)
    elif(image_name.endswith(".jpg") or image_name.endswith(".jpeg") or
         image_name.endswith(".tif") or image_name.endswith(".png")):
        image_dict = load_rgb_image_as_3d_array(image_name)
    else:
        raise ValueError("unsupported image format")
    return image_dict

def save_array_as_nifty_volume(data, image_name, reference_name = None):
    """
    Save a numpy array as nifty image

    :param data:  (numpy.ndarray) A numpy array with shape [Depth, Height, Width].
    :param image_name:  (str) The ouput file name.
    :param reference_name:  (str) File name of the reference image of which 
        meta information is used.
    """
    img = sitk.GetImageFromArray(data)
    if(reference_name is not None):
        img_ref = sitk.ReadImage(reference_name)
        #img.CopyInformation(img_ref)
        img.SetSpacing(img_ref.GetSpacing())
        img.SetOrigin(img_ref.GetOrigin())
        img.SetDirection(img_ref.GetDirection())
    sitk.WriteImage(img, image_name)

def save_array_as_rgb_image(data, image_name):
    """
    Save a numpy array as rgb image.

    :param data:  (numpy.ndarray) A numpy array with shape [3, H, W] or
        [H, W, 3] or [H, W]. 
    :param image_name:  (str) The output file name.
    """
    data_dim = len(data.shape)
    if(data_dim == 3):
        assert(data.shape[0] == 3 or data.shape[2] == 3)
        if(data.shape[0] == 3):
            data = np.transpose(data, [1, 2, 0])
    img = Image.fromarray(data)
    img.save(image_name)

def save_nd_array_as_image(data, image_name, reference_name = None):
    """
    Save a 3D or 2D numpy array as medical image or RGB image
    
    :param data:  (numpy.ndarray) A numpy array with shape [3, H, W] or
        [H, W, 3] or [H, W]. 
    :param reference_name: (str) File name of the reference image of which 
        meta information is used.
    """
    data_dim = len(data.shape)
    assert(data_dim == 2 or data_dim == 3)
    if (image_name.endswith(".nii.gz") or image_name.endswith(".nii") or
        image_name.endswith(".mha")):
        assert(data_dim == 3)
        save_array_as_nifty_volume(data, image_name, reference_name)

    elif(image_name.endswith(".jpg") or image_name.endswith(".jpeg") or
         image_name.endswith(".tif") or image_name.endswith(".png")):
         assert(data_dim == 2)
         save_array_as_rgb_image(data, image_name)
    else:
        raise ValueError("unsupported image format {0:}".format(
            image_name.split('.')[-1]))

def rotate_nifty_volume_to_LPS(filename_or_image_dict, origin = None, direction = None):
    '''
    Rotate the axis of a 3D volume to LPS

    :param filename_or_image_dict: (str) Filename of the nifty file (str) or image dictionary 
        returned by load_nifty_volume_as_4d_array. If supplied with the former, 
        the flipped image data will be saved to override the original file. 
        If supplied with the later, only flipped image data will be returned.\n
    :param origin: (list/tuple) The origin of the image.
    :param direction: (list or tuple) The direction of the image.

    :return: A dictionary for image data and meta info, with ``data_array``,
        ``origin``, ``direction`` and ``spacing``.
    '''

    if type(filename_or_image_dict) == str:
        image_data = load_nifty_volume_as_4d_array(filename_or_image_dict)
        save_nifty = True
    elif type(filename_or_image_dict) == dict:
        image_data = filename_or_image_dict
        save_nifty = False

    data_array = image_data['data_array']
    if not origin:
        origin = image_data['origin']
    if not direction:
        direction = image_data['direction']
    spacing = image_data['spacing']

    fliped = False
    if direction[0] == -1.:
        data_array = np.flip(data_array, axis = 3)
        fliped = True
    if direction[4] == -1.:
        data_array = np.flip(data_array, axis = 2)
        fliped = True
    if direction[8] == -1.:
        data_array = np.flip(data_array, axis = 1)
        fliped = True

    if save_nifty:
        if not fliped:
            return
        else:
            print(f'rotate {filename_or_image_dict} to LPS')
            img = sitk.GetImageFromArray(data_array[0])
            img.SetSpacing(spacing)
            img.SetOrigin(origin)
            img.SetDirection([1., 0., 0., 0., 1., 0., 0., 0., 1.])
            sitk.WriteImage(img, filename_or_image_dict)
    else:
        image_data['data_array'] = data_array
        image_data['direction'] = [1., 0., 0., 0., 1., 0., 0., 0., 1.]
        return image_data



# -*- coding: utf-8 -*-
"""
code adapted from: https://github.com/HuCaoFighting/Swin-Unet

"""
from __future__ import print_function, division

import copy
import numpy as np 
import torch
import torch.nn as nn

from pymic.net.net2d.trans2d.swinunet_sys import SwinTransformerSys

class SwinUNet(nn.Module):
    """
    Note: the input size should be 224x224
    """
    def __init__(self, params):
        super(SwinUNet, self).__init__()
        img_size    = params['img_size']
        if(isinstance(img_size, tuple) or isinstance(img_size, list)):
            img_size = img_size[0]
        self.num_classes = params['class_num']
        self.swin_unet = SwinTransformerSys(img_size = img_size, num_classes=self.num_classes)
        # self.swin_unet = SwinTransformerSys(img_size=config.DATA.IMG_SIZE,
        #                         patch_size=config.MODEL.SWIN.PATCH_SIZE,
        #                         in_chans=config.MODEL.SWIN.IN_CHANS,
        #                         num_classes=self.num_classes,
        #                         embed_dim=config.MODEL.SWIN.EMBED_DIM,
        #                         depths=config.MODEL.SWIN.DEPTHS,
        #                         num_heads=config.MODEL.SWIN.NUM_HEADS,
        #                         window_size=config.MODEL.SWIN.WINDOW_SIZE,
        #                         mlp_ratio=config.MODEL.SWIN.MLP_RATIO,
        #                         qkv_bias=config.MODEL.SWIN.QKV_BIAS,
        #                         qk_scale=config.MODEL.SWIN.QK_SCALE,
        #                         drop_rate=config.MODEL.DROP_RATE,
        #                         drop_path_rate=config.MODEL.DROP_PATH_RATE,
        #                         ape=config.MODEL.SWIN.APE,
        #                         patch_norm=config.MODEL.SWIN.PATCH_NORM,
        #                         use_checkpoint=config.TRAIN.USE_CHECKPOINT)

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)

        if x.size()[1] == 1:
            x = x.repeat(1,3,1,1)
        logits = self.swin_unet(x)

        if(len(x_shape) == 5):
            new_shape = [N, D] + list(logits.shape)[1:]
            logits = torch.reshape(logits, new_shape)
            logits = torch.transpose(logits, 1, 2)

        return logits

    def load_from(self, config):
        pretrained_path = config.MODEL.PRETRAIN_CKPT
        if pretrained_path is not None:
            print("pretrained_path:{}".format(pretrained_path))
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
            pretrained_dict = torch.load(pretrained_path, map_location=device)
            if "model"  not in pretrained_dict:
                print("---start load pretrained modle by splitting---")
                pretrained_dict = {k[17:]:v for k,v in pretrained_dict.items()}
                for k in list(pretrained_dict.keys()):
                    if "output" in k:
                        print("delete key:{}".format(k))
                        del pretrained_dict[k]
                msg = self.swin_unet.load_state_dict(pretrained_dict,strict=False)
                # print(msg)
                return
            pretrained_dict = pretrained_dict['model']
            print("---start load pretrained modle of swin encoder---")

            model_dict = self.swin_unet.state_dict()
            full_dict = copy.deepcopy(pretrained_dict)
            for k, v in pretrained_dict.items():
                if "layers." in k:
                    current_layer_num = 3-int(k[7:8])
                    current_k = "layers_up." + str(current_layer_num) + k[8:]
                    full_dict.update({current_k:v})
            for k in list(full_dict.keys()):
                if k in model_dict:
                    if full_dict[k].shape != model_dict[k].shape:
                        print("delete:{};shape pretrain:{};shape model:{}".format(k,v.shape,model_dict[k].shape))
                        del full_dict[k]

            msg = self.swin_unet.load_state_dict(full_dict, strict=False)
            # print(msg)
        else:
            print("none pretrain")

        
if __name__ == "__main__":
    params = {'img_size': [224, 224],
              'class_num': 2}
    net = SwinUNet(params)
    net.double()

    x  = np.random.rand(4, 1, 224, 224)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = net(xt)
    print(len(y.size()))
    y = y.detach().numpy()
    print(y.shape)


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import math
import torch
import torch.nn as nn
import numpy as np
from torch.nn import Dropout, Softmax, Linear, Conv2d, LayerNorm
from pymic.net.net3d.unet3d import Encoder, Decoder

class Attention(nn.Module):
    def __init__(self, params):
        super(Attention, self).__init__()
        hidden_size = params["attention_hidden_size"]
        self.num_attention_heads = params["attention_num_heads"]
        self.attention_head_size = int(hidden_size / self.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = Linear(hidden_size, self.all_head_size)
        self.key   = Linear(hidden_size, self.all_head_size)
        self.value = Linear(hidden_size, self.all_head_size)

        self.out = Linear(hidden_size, hidden_size)
        self.attn_dropout = Dropout(params["attention_dropout_rate"])
        self.proj_dropout = Dropout(params["attention_dropout_rate"])

        self.softmax = Softmax(dim=-1)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        attention_probs = self.softmax(attention_scores)
        # weights = attention_probs if self.vis else None
        attention_probs = self.attn_dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        attention_output = self.out(context_layer)
        attention_output = self.proj_dropout(attention_output)
        return attention_output

class MLP(nn.Module):
    def __init__(self, params):
        super(MLP, self).__init__()
        hidden_size = params["attention_hidden_size"]
        mlp_dim     = params["attention_mlp_dim"]
        self.fc1 = Linear(hidden_size, mlp_dim)
        self.fc2 = Linear(mlp_dim, hidden_size)
        self.act_fn = torch.nn.functional.gelu
        self.dropout = Dropout(params["attention_dropout_rate"])

        self._init_weights()

    def _init_weights(self):
        nn.init.xavier_uniform_(self.fc1.weight)
        nn.init.xavier_uniform_(self.fc2.weight)
        nn.init.normal_(self.fc1.bias, std=1e-6)
        nn.init.normal_(self.fc2.bias, std=1e-6)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act_fn(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.dropout(x)
        return x

class Block(nn.Module):
    def __init__(self, params):
        super(Block, self).__init__()
        hidden_size = params["attention_hidden_size"]
        self.attention_norm = LayerNorm(hidden_size, eps=1e-6)
        self.ffn_norm = LayerNorm(hidden_size, eps=1e-6)
        self.ffn = MLP(params)
        self.attn = Attention(params)

    def forward(self, x):
        # convert the tensor shape from [B, C, D, H, W] to [B, DHW, C]
        [B, C, D, H, W] = list(x.shape)
        new_shape = [B, C, D*H*W]
        x = torch.reshape(x, new_shape)
        x = torch.transpose(x, 1, 2)

        h = x
        x = self.attention_norm(x)
        x = self.attn(x)
        x = x + h

        h = x
        x = self.ffn_norm(x)
        x = self.ffn(x)
        x = x + h

        # convert the result back to [B, C, D, H, W]
        x = torch.transpose(x, 1, 2)
        x = torch.reshape(x, [B, C, D, H, W])
        return x

class MedFormerV1(nn.Module):
    """
    An implementation of the U-Net.
        
    * Reference: Özgün Çiçek, Ahmed Abdulkadir, Soeren S. Lienkamp, Thomas Brox, Olaf Ronneberger:
      3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation. 
      `MICCAI (2) 2016: 424-432. <https://link.springer.com/chapter/10.1007/978-3-319-46723-8_49>`_
    
    Note that there are some modifications from the original paper, such as
    the use of batch normalization, dropout, leaky relu and deep supervision.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using trilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    :param deep_supervise: (bool) Using deep supervision for training or not.
    """
    def __init__(self, params):
        super(MedFormerV1, self).__init__()
        self.params   = params
        self.encoder  = Encoder(params)
        self.decoder  = Decoder(params)  
        self.attn = Block(params)

    def forward(self, x):
        f = self.encoder(x)
        f[-1] = self.attn(f[-1])
        output = self.decoder(f)
        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'class_num': 2,
              'feature_chns':[16, 32, 64, 128],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True,
              'deep_supervise': True,
              'attention_hidden_size': 128,
              'attention_num_heads': 4,
              'attention_mlp_dim': 256,
              'attention_dropout_rate': 0.2}
    Net = MedFormerV1(params)
    Net = Net.double()

    x  = np.random.rand(1, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print("output length", len(y))
    for yi in y:
        yi = yi.detach().numpy()
        print(yi.shape)
